import math

import numpy as np
import pytest
import tensorflow as tf
from tensorflow.python.keras import testing_utils

import larq as lq
from larq import testing_utils as lq_testing_utils
from larq.callbacks import HyperparameterScheduler


class TestHyperparameterScheduler:
    def _create_data_and_model(self, train_samples=1000):
        np.random.seed(1337)
        (x_train, y_train), (x_test, y_test) = testing_utils.get_test_data(
            train_samples=train_samples,
            test_samples=0,
            input_shape=(10,),
            num_classes=2,
        )
        y_train = tf.keras.utils.to_categorical(y_train)

        model = lq_testing_utils.get_small_bnn_model(
            x_train.shape[1], 20, y_train.shape[1]
        )

        return x_train, y_train, model

    def test_normal_optimizer(self):
        x_train, y_train, model = self._create_data_and_model()

        model.compile(
            loss="categorical_crossentropy",
            optimizer=tf.keras.optimizers.Adam(0.01),
            metrics=["accuracy"],
        )

        def scheduler(x):
            return 1.0 / (1.0 + x)

        # We shouldn' t need to specify the optimizer
        test_scheduler = HyperparameterScheduler(
            schedule=scheduler,
            hyperparameter="lr",
            verbose=1,
        )

        num_epochs = 2
        model.fit(
            x_train,
            y_train,
            epochs=num_epochs,
            batch_size=16,
            callbacks=[test_scheduler],
            verbose=0,
        )

        np.testing.assert_almost_equal(
            tf.keras.backend.get_value(model.optimizer.lr),
            scheduler(num_epochs - 1),
            decimal=8,
        )

    def test_per_step(self):
        train_samples = 20
        x_train, y_train, model = self._create_data_and_model(train_samples)

        model.compile(
            loss="categorical_crossentropy",
            optimizer=tf.keras.optimizers.Adam(0.01),
            metrics=["accuracy"],
        )

        def scheduler(x):
            return 1.0 / (1.0 + x)

        # Test that we don't accept incorrect `update_freq`
        with pytest.raises(ValueError):
            HyperparameterScheduler(
                schedule=scheduler,
                hyperparameter="lr",
                update_freq="wrong",
            )

        # The actual scheduler we'll use
        test_scheduler = HyperparameterScheduler(
            schedule=scheduler,
            hyperparameter="lr",
            update_freq="step",
            verbose=1,
        )

        num_epochs = 1
        batch_size = 10
        model.fit(
            x_train,
            y_train,
            epochs=num_epochs,
            batch_size=16,
            callbacks=[test_scheduler],
            verbose=0,
        )

        np.testing.assert_almost_equal(
            tf.keras.backend.get_value(model.optimizer.lr),
            scheduler(math.ceil(train_samples / batch_size) - 1),
            decimal=8,
        )

    def test_case_optimizer(self):
        x_train, y_train, model = self._create_data_and_model()

        bop = lq.optimizers.Bop(threshold=1e-6, gamma=1e-3)
        adam = tf.keras.optimizers.Adam(0.01)
        case_optimizer = lq.optimizers.CaseOptimizer(
            (lq.optimizers.Bop.is_binary_variable, bop),
            default_optimizer=adam,
        )

        model.compile(
            loss="categorical_crossentropy",
            optimizer=case_optimizer,
            metrics=["accuracy"],
        )

        def scheduler(x):
            return 1.0 / (1.0 + x)

        cbk_gamma_scheduler = HyperparameterScheduler(
            schedule=scheduler,
            optimizer=model.optimizer.optimizers[0],
            hyperparameter="gamma",
            verbose=1,
        )
        cbk_threshold_scheduler = HyperparameterScheduler(
            schedule=scheduler,
            optimizer=model.optimizer.optimizers[0],
            hyperparameter="threshold",
            verbose=1,
        )
        cbk_lr_scheduler = HyperparameterScheduler(
            schedule=scheduler,
            optimizer=model.optimizer.optimizers[1],
            hyperparameter="lr",
            verbose=1,
        )

        num_epochs = 3
        model.fit(
            x_train,
            y_train,
            epochs=num_epochs,
            batch_size=16,
            callbacks=[cbk_gamma_scheduler, cbk_lr_scheduler, cbk_threshold_scheduler],
            verbose=0,
        )

        np.testing.assert_almost_equal(
            tf.keras.backend.get_value(model.optimizer.optimizers[0].gamma),
            scheduler(num_epochs - 1),
            decimal=8,
        )

        np.testing.assert_almost_equal(
            tf.keras.backend.get_value(model.optimizer.optimizers[0].threshold),
            scheduler(num_epochs - 1),
            decimal=8,
        )

        np.testing.assert_almost_equal(
            tf.keras.backend.get_value(model.optimizer.optimizers[1].lr),
            scheduler(num_epochs - 1),
            decimal=8,
        )

    def test_wrong_param(self):
        x_train, y_train, model = self._create_data_and_model()

        model.compile(
            loss="categorical_crossentropy",
            optimizer=tf.keras.optimizers.Adam(0.01),
            metrics=["accuracy"],
        )

        def scheduler(x):
            return 1.0 / (1.0 + x)

        wrong_scheduler = HyperparameterScheduler(
            schedule=scheduler,
            hyperparameter="invalid_param",
            verbose=1,
        )

        with pytest.raises(ValueError):
            model.fit(
                x_train,
                y_train,
                epochs=1,
                batch_size=16,
                callbacks=[wrong_scheduler],
                verbose=0,
            )



"""Math operations that are specific to extremely quantized networks."""

import tensorflow as tf


def sign(x):
    r"""A sign function that will never be zero
    \\[
    f(x) = \begin{cases}
      -1 & x < 0 \\\
      \hphantom{-}1 & x \geq 0
    \end{cases}
    \\]

    This function is similar to
    [`tf.math.sign`](https://www.tensorflow.org/api_docs/python/tf/math/sign) but will
    return a binary value and will never be zero.

    # Arguments
        `x`: Input Tensor

    # Returns
        A Tensor with same type as `x`.
    """
    return tf.sign(tf.sign(x) + 0.1)


def heaviside(x):
    r"""Heaviside step function with output values 0 and 1.

    \\[
    q(x) = \begin{cases}
    +1 & x > 0 \\\
    \hphantom{+}0 & x \leq 0
    \end{cases}
    \\]

    # Arguments
        `x`: Input Tensor

    # Returns
        A Tensor with same type as `x`.
    """
    return tf.sign(tf.nn.relu(x))



import inspect

import numpy as np
import pytest
import tensorflow as tf

import larq as lq
from larq import testing_utils

PARAMS_ALL_LAYERS = [
    (lq.layers.QuantDense, tf.keras.layers.Dense, (3, 2), dict(units=3)),
    (
        lq.layers.QuantConv1D,
        tf.keras.layers.Conv1D,
        (2, 3, 7),
        dict(filters=2, kernel_size=3),
    ),
    (
        lq.layers.QuantConv2D,
        tf.keras.layers.Conv2D,
        (2, 3, 7, 6),
        dict(filters=2, kernel_size=3),
    ),
    (
        lq.layers.QuantConv3D,
        tf.keras.layers.Conv3D,
        (2, 3, 7, 6, 5),
        dict(filters=2, kernel_size=3),
    ),
    (
        lq.layers.QuantConv2DTranspose,
        tf.keras.layers.Conv2DTranspose,
        (2, 3, 7, 6),
        dict(filters=2, kernel_size=3),
    ),
    (
        lq.layers.QuantConv3DTranspose,
        tf.keras.layers.Conv3DTranspose,
        (2, 3, 7, 6, 5),
        dict(filters=2, kernel_size=3),
    ),
    (
        lq.layers.QuantLocallyConnected1D,
        tf.keras.layers.LocallyConnected1D,
        (2, 8, 5),
        dict(filters=4, kernel_size=3),
    ),
    (
        lq.layers.QuantLocallyConnected2D,
        tf.keras.layers.LocallyConnected2D,
        (8, 6, 10, 4),
        dict(filters=3, kernel_size=3),
    ),
]

PARAMS_SEP_LAYERS = [
    (lq.layers.QuantSeparableConv1D, tf.keras.layers.SeparableConv1D, (2, 3, 7)),
    (lq.layers.QuantSeparableConv2D, tf.keras.layers.SeparableConv2D, (2, 3, 7, 6)),
]


class TestLayers:
    @pytest.mark.parametrize(
        "quantized_layer, layer, input_shape, kwargs", PARAMS_ALL_LAYERS
    )
    def test_binarization(
        self, quantized_layer, layer, input_shape, kwargs, keras_should_run_eagerly
    ):
        input_data = testing_utils.random_input(input_shape)
        random_weight = np.random.random() - 0.5

        with lq.context.metrics_scope(["flip_ratio"]):
            quant_output = testing_utils.layer_test(
                quantized_layer,
                kwargs=dict(
                    **kwargs,
                    kernel_quantizer="ste_sign",
                    input_quantizer="ste_sign",
                    kernel_initializer=tf.keras.initializers.constant(random_weight),
                ),
                input_data=input_data,
                should_run_eagerly=keras_should_run_eagerly,
            )

        fp_model = tf.keras.models.Sequential(
            [
                layer(
                    **kwargs,
                    kernel_initializer=tf.keras.initializers.constant(
                        np.sign(random_weight)
                    ),
                    input_shape=input_shape[1:],
                )
            ]
        )

        np.testing.assert_allclose(quant_output, fp_model.predict(np.sign(input_data)))

    @pytest.mark.parametrize("quantized_layer, layer, input_shape", PARAMS_SEP_LAYERS)
    def test_separable_layers(
        self, quantized_layer, layer, input_shape, keras_should_run_eagerly
    ):
        input_data = testing_utils.random_input(input_shape)
        random_d_kernel = np.random.random() - 0.5
        random_p_kernel = np.random.random() - 0.5

        with lq.context.metrics_scope(["flip_ratio"]):
            quant_output = testing_utils.layer_test(
                quantized_layer,
                kwargs=dict(
                    filters=3,
                    kernel_size=3,
                    depthwise_quantizer="ste_sign",
                    pointwise_quantizer="ste_sign",
                    input_quantizer="ste_sign",
                    depthwise_initializer=tf.keras.initializers.constant(
                        random_d_kernel
                    ),
                    pointwise_initializer=tf.keras.initializers.constant(
                        random_p_kernel
                    ),
                ),
                input_data=input_data,
                should_run_eagerly=keras_should_run_eagerly,
            )

        fp_model = tf.keras.models.Sequential(
            [
                layer(
                    filters=3,
                    kernel_size=3,
                    depthwise_initializer=tf.keras.initializers.constant(
                        np.sign(random_d_kernel)
                    ),
                    pointwise_initializer=tf.keras.initializers.constant(
                        np.sign(random_p_kernel)
                    ),
                    input_shape=input_shape[1:],
                )
            ]
        )

        np.testing.assert_allclose(quant_output, fp_model.predict(np.sign(input_data)))

    def test_depthwise_layers(self, keras_should_run_eagerly):
        input_data = testing_utils.random_input((2, 3, 7, 6))
        random_weight = np.random.random() - 0.5

        with lq.context.metrics_scope(["flip_ratio"]):
            quant_output = testing_utils.layer_test(
                lq.layers.QuantDepthwiseConv2D,
                kwargs=dict(
                    kernel_size=3,
                    depthwise_quantizer="ste_sign",
                    input_quantizer="ste_sign",
                    depthwise_initializer=tf.keras.initializers.constant(random_weight),
                ),
                input_data=input_data,
                should_run_eagerly=keras_should_run_eagerly,
            )

        fp_model = tf.keras.models.Sequential(
            [
                tf.keras.layers.DepthwiseConv2D(
                    kernel_size=3,
                    depthwise_initializer=tf.keras.initializers.constant(
                        np.sign(random_weight)
                    ),
                    input_shape=input_data.shape[1:],
                )
            ]
        )

        np.testing.assert_allclose(quant_output, fp_model.predict(np.sign(input_data)))

    @pytest.mark.parametrize(
        "layer_cls, input_dim",
        [
            (lq.layers.QuantConv1D, 3),
            (lq.layers.QuantConv2D, 4),
            (lq.layers.QuantConv3D, 5),
            (lq.layers.QuantSeparableConv1D, 3),
            (lq.layers.QuantSeparableConv2D, 4),
            (lq.layers.QuantDepthwiseConv2D, 4),
        ],
    )
    @pytest.mark.parametrize("dilation", [True, False])
    def test_non_zero_padding_layers(
        self, mocker, layer_cls, input_dim, data_format, dilation
    ):
        inputs = np.zeros(np.random.randint(5, 20, size=input_dim), np.float32)
        kernel = tuple(np.random.randint(3, 7, size=input_dim - 2))
        rand_tuple = tuple(np.random.randint(1, 4, size=input_dim - 2))
        if not dilation and layer_cls in (
            lq.layers.QuantSeparableConv2D,
            lq.layers.QuantDepthwiseConv2D,
        ):
            rand_tuple = int(rand_tuple[0])
        kwargs = {"dilation_rate": rand_tuple} if dilation else {"strides": rand_tuple}

        args = (kernel,) if layer_cls == lq.layers.QuantDepthwiseConv2D else (2, kernel)
        ref_layer = layer_cls(*args, padding="same", **kwargs)
        spy = mocker.spy(tf, "pad")
        layer = layer_cls(*args, padding="same", pad_values=1.0, **kwargs)
        layer.build(inputs.shape)
        conv_op = getattr(layer, "_convolution_op", None)
        assert layer(inputs).shape == ref_layer(inputs).shape
        spy.assert_called_once_with(mocker.ANY, mocker.ANY, constant_values=1.0)
        assert conv_op == getattr(layer, "_convolution_op", None)

    @pytest.mark.parametrize(
        "layer_cls",
        [
            lq.layers.QuantConv1D,
            lq.layers.QuantConv2D,
            lq.layers.QuantConv3D,
            lq.layers.QuantSeparableConv1D,
            lq.layers.QuantSeparableConv2D,
            lq.layers.QuantDepthwiseConv2D,
        ],
    )
    @pytest.mark.parametrize("static", [True, False])
    def test_non_zero_padding_shapes(self, layer_cls, data_format, static):
        layer = layer_cls(
            16, 3, padding="same", pad_values=1.0, data_format=data_format
        )
        input_shape = [32 if static else None] * layer.rank + [3]
        if data_format == "channels_first":
            input_shape = reversed(input_shape)
        input = tf.keras.layers.Input(shape=input_shape)

        layer(input)
        if static:
            for dim in layer.output_shape[1:]:
                assert dim is not None


class TestLayerWarns:
    def test_layer_warns(self, caplog):
        lq.layers.QuantDense(5, kernel_quantizer="ste_sign")
        assert len(caplog.records) >= 1
        assert "kernel_constraint" in caplog.text

    def test_layer_does_not_warn(self, caplog):
        lq.layers.QuantDense(
            5, kernel_quantizer="ste_sign", kernel_constraint="weight_clip"
        )
        assert "kernel_constraint" not in caplog.text

    def test_depthwise_layer_warns(self, caplog):
        lq.layers.QuantDepthwiseConv2D(5, depthwise_quantizer="ste_sign")
        assert len(caplog.records) >= 1
        assert "depthwise_constraint" in caplog.text

    def test_depthwise_layer_does_not_warn(self, caplog):
        lq.layers.QuantDepthwiseConv2D(
            5, depthwise_quantizer="ste_sign", depthwise_constraint="weight_clip"
        )
        assert "depthwise_constraint" not in caplog.text

    def test_separable_layer_warns(self, caplog):
        lq.layers.QuantSeparableConv2D(
            3, 3, depthwise_quantizer="ste_sign", pointwise_quantizer="ste_sign"
        )
        assert "depthwise_constraint" in caplog.text
        assert "pointwise_constraint" in caplog.text

    def test_separable_layer_does_not_warn(self, caplog):
        lq.layers.QuantSeparableConv2D(
            3,
            3,
            depthwise_quantizer="ste_sign",
            pointwise_quantizer="ste_sign",
            depthwise_constraint="weight_clip",
            pointwise_constraint="weight_clip",
        )
        assert caplog.records == []

    def test_conv1d_non_zero_padding_raises(self):
        with pytest.raises(ValueError, match=r".*pad_values.*"):
            lq.layers.QuantConv1D(24, 3, padding="causal", pad_values=1.0)


@pytest.mark.parametrize(
    "quant_layer,layer",
    [
        (lq.layers.QuantDense, tf.keras.layers.Dense),
        (lq.layers.QuantConv1D, tf.keras.layers.Conv1D),
        (lq.layers.QuantConv2D, tf.keras.layers.Conv2D),
        (lq.layers.QuantConv3D, tf.keras.layers.Conv3D),
        (lq.layers.QuantConv2DTranspose, tf.keras.layers.Conv2DTranspose),
        (lq.layers.QuantConv3DTranspose, tf.keras.layers.Conv3DTranspose),
        (lq.layers.QuantLocallyConnected1D, tf.keras.layers.LocallyConnected1D),
        (lq.layers.QuantLocallyConnected2D, tf.keras.layers.LocallyConnected2D),
        (lq.layers.QuantDepthwiseConv2D, tf.keras.layers.DepthwiseConv2D),
    ],
)
def test_layer_kwargs(quant_layer, layer):
    quant_params = inspect.signature(quant_layer).parameters
    params = inspect.signature(layer).parameters

    quant_params_list = list(quant_params.keys())
    params_list = list(params.keys())

    for p in (
        "input_quantizer",
        "kernel_quantizer",
        "depthwise_quantizer",
        "pointwise_quantizer",
        "pad_values",
    ):
        try:
            quant_params_list.remove(p)
        except ValueError:
            pass
    assert quant_params_list == params_list

    for param in params_list:
        assert quant_params.get(param).default == params.get(param).default  # type: ignore



import pytest
import tensorflow as tf
from packaging import version
from tensorflow.python.distribute import strategy_combinations
from tensorflow.python.eager import context

from larq import context as lq_context

if version.parse(tf.__version__) >= version.parse("1.15"):
    strategy_combinations.set_virtual_cpus_to_at_least(3)
    distributed_devices = ["/cpu:1", "/cpu:2"]
else:
    distributed_devices = ["/cpu:0"]


@pytest.fixture
def eager_mode():
    """pytest fixture for running test in eager mode"""
    with context.eager_mode():
        yield


@pytest.fixture
def graph_mode():
    """pytest fixture for running test in graph mode"""
    with context.graph_mode():
        with tf.compat.v1.Session().as_default():
            yield
            tf.keras.backend.clear_session()


@pytest.fixture(params=["eager", "graph"])
def eager_and_graph_mode(request):
    """pytest fixture for running test in eager and graph mode"""
    if request.param == "graph":
        with context.graph_mode():
            with tf.compat.v1.Session().as_default():
                yield request.param
                tf.keras.backend.clear_session()
    else:
        with context.eager_mode():
            yield request.param


@pytest.fixture(params=["graph", "tf_eager", "tf_keras_eager"])
def keras_should_run_eagerly(request):
    """Fixture to run in graph and two eager modes.

    The modes are:
    - Graph mode
    - TensorFlow eager and Keras eager
    - TensorFlow eager and Keras not eager

    The `tf.context` sets graph/eager mode for TensorFlow. The yield is True if Keras
    should run eagerly.
    """

    if request.param == "graph":
        if version.parse(tf.__version__) >= version.parse("2"):
            pytest.skip("Skipping graph mode for TensorFlow 2+.")

        with context.graph_mode():
            yield
    else:
        with context.eager_mode():
            yield request.param == "tf_keras_eager"


@pytest.fixture(params=[False, True])
def distribute_scope(request):
    if request.param is True:
        with tf.distribute.MirroredStrategy(distributed_devices).scope():
            yield request.param
    else:
        yield request.param


@pytest.fixture(params=[True, False])
def quantized(request):
    """pytest fixture for running test quantized and non-quantized"""
    with lq_context.quantized_scope(request.param):
        yield request.param


@pytest.fixture(params=["channels_last", "channels_first"])
def data_format(request):
    return request.param



from setuptools import find_packages, setup


def readme():
    with open("README.md", "r") as f:
        return f.read()


setup(
    name="larq",
    version="0.11.0",
    python_requires=">=3.6",
    author="Plumerai",
    author_email="opensource@plumerai.com",
    description="An Open Source Machine Learning Library for Training Binarized Neural Networks",
    long_description=readme(),
    long_description_content_type="text/markdown",
    url="https://larq.dev/",
    packages=find_packages(exclude=["larq.snapshots"]),
    license="Apache 2.0",
    install_requires=[
        "numpy >= 1.15.4, < 2.0",
        "terminaltables>=3.1.0",
        "dataclasses ; python_version<'3.7'",
        "importlib-metadata >= 2.0, < 4.0 ; python_version<'3.8'",
    ],
    extras_require={
        "tensorflow": ["tensorflow>=1.14.0"],
        "tensorflow_gpu": ["tensorflow-gpu>=1.14.0"],
        "test": [
            "black==20.8b1",
            "flake8>=3.7.9,<3.9.0",
            "isort==5.7.0",
            "packaging>=19.2,<21.0",
            "pytest>=5.2.4,<6.3.0",
            "pytest-cov>=2.8.1,<2.12.0",
            "pytest-xdist>=1.30,<2.3",
            "pytest-mock>=2.0,<3.6",
            "pytype==2020.10.8",
            "snapshottest>=0.5.1,<0.7.0",
        ],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



"""We add metrics specific to extremely quantized networks using a
`larq.context.metrics_scope` rather than through the `metrics` parameter of
`model.compile()`, where most common metrics reside. This is because, to calculate
metrics like the `flip_ratio`, we need a layer's kernel or activation and not just the
`y_true` and `y_pred` that Keras passes to metrics defined in the usual way.
"""

import numpy as np
import tensorflow as tf

from larq import utils


@utils.register_alias("flip_ratio")
@utils.register_keras_custom_object
class FlipRatio(tf.keras.metrics.Metric):
    """Computes the mean ratio of changed values in a given tensor.

    !!! example
        ```python
        m = metrics.FlipRatio()
        m.update_state((1, 1))  # result: 0
        m.update_state((2, 2))  # result: 1
        m.update_state((1, 2))  # result: 0.75
        print('Final result: ', m.result().numpy())  # Final result: 0.75
        ```

    # Arguments
        name: Name of the metric.
        values_dtype: Data type of the tensor for which to track changes.
        dtype: Data type of the moving mean.
    """

    def __init__(self, values_dtype="int8", name="flip_ratio", dtype=None):
        super().__init__(name=name, dtype=dtype)
        self.built = False
        self.values_dtype = tf.as_dtype(values_dtype)

    def build(self, input_shape):
        self._previous_values = self.add_weight(
            "previous_values",
            shape=input_shape,
            dtype=self.values_dtype,
            initializer=tf.keras.initializers.zeros,
            aggregation=tf.VariableAggregation.ONLY_FIRST_REPLICA,
        )
        self.total = self.add_weight(
            "total",
            initializer=tf.keras.initializers.zeros,
            aggregation=tf.VariableAggregation.ONLY_FIRST_REPLICA,
        )
        self.count = self.add_weight(
            "count",
            initializer=tf.keras.initializers.zeros,
            aggregation=tf.VariableAggregation.ONLY_FIRST_REPLICA,
        )
        self._size = tf.cast(np.prod(input_shape), self.dtype)
        self.built = True

    def update_state(self, values, sample_weight=None):
        values = tf.cast(values, self.values_dtype)

        if not self.built:
            with tf.name_scope(self.name), tf.init_scope():
                self.build(values.shape)

        unchanged_values = tf.math.count_nonzero(
            tf.equal(self._previous_values, values)
        )
        flip_ratio = 1 - (
            tf.cast(unchanged_values, self.dtype) / tf.cast(self._size, self.dtype)
        )

        update_total_op = self.total.assign_add(flip_ratio * tf.sign(self.count))
        with tf.control_dependencies([update_total_op]):
            update_count_op = self.count.assign_add(1)
            with tf.control_dependencies([update_count_op]):
                return self._previous_values.assign(values)

    def result(self):
        return tf.compat.v1.div_no_nan(self.total, self.count - 1)

    def reset_states(self):
        tf.keras.backend.batch_set_value(
            [(v, 0) for v in self.variables if v is not self._previous_values]
        )

    def get_config(self):
        return {**super().get_config(), "values_dtype": self.values_dtype.name}



import pytest

from larq import context


def test_scope():
    assert context.get_training_metrics() == set()
    with context.metrics_scope(["flip_ratio"]):
        assert context.get_training_metrics() == {"flip_ratio"}
    assert context.get_training_metrics() == set()
    with pytest.raises(ValueError, match=r".*unknown_metric.*"):
        with context.metrics_scope(["flip_ratio", "unknown_metric"]):
            pass
